unit UnitSystemCal;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, ComCtrls, Grids, ExtCtrls;

type
  TMainForm = class(TForm)
    NextButton: TButton;
    CancelBtn: TButton;
    PrevButton: TButton;
    MainPages: TPageControl;
    OverviewSheet: TTabSheet;
    OverviewInstructionsMemo: TMemo;
    OffsetSheet: TTabSheet;
    Label1: TLabel;
    Label2: TLabel;
    OffsetInstructionsLabel: TLabel;
    OffsetExpectedEdit: TEdit;
    OffsetObservedEdit: TEdit;
    OffsetInstructionsMemo: TMemo;
    GainSheet: TTabSheet;
    Label7: TLabel;
    Label8: TLabel;
    GainInstructionsLabel: TLabel;
    GainExpectedEdit: TEdit;
    GainObservedEdit: TEdit;
    GainInstructionsMemo: TMemo;
    ConfirmSheet: TTabSheet;
    SaveEm: TSaveDialog;
    Label5: TLabel;
    SaveSheet: TTabSheet;
    CalFileEdit: TEdit;
    BrowseButton: TButton;
    Label6: TLabel;
    ConfirmObservedEdit: TEdit;
    Label9: TLabel;
    Label10: TLabel;
    Label11: TLabel;
    Label12: TLabel;
    Label13: TLabel;
    SaveInstructionsMemo: TMemo;
    RangeSheet: TTabSheet;
    AcqTime: TTimer;
    RangeList: TListBox;
    RangeInstructionsMemo: TMemo;
    ScrollBox1: TScrollBox;
    CalPaint: TPaintBox;
    SERadio: TRadioButton;
    DifRadio: TRadioButton;
    FuncMemo: TMemo;
    procedure FormCreate(Sender: TObject);
    procedure CancelBtnClick(Sender: TObject);
    procedure PrevButtonClick(Sender: TObject);
    procedure NextButtonClick(Sender: TObject);
    procedure MainPagesChange(Sender: TObject);
    procedure MainPagesChanging(Sender: TObject; var AllowChange: Boolean);
    procedure BrowseButtonClick(Sender: TObject);
    procedure AcqTimeTimer(Sender: TObject);
    procedure RangeListClick(Sender: TObject);
    procedure CalPaintPaint(Sender: TObject);
    procedure SEDifRadioClick(Sender: TObject);
  private
    { Private declarations }
  public
    { Public declarations }
  end;

var
  MainForm: TMainForm;

implementation

{$R *.DFM}

uses
  Math, AIOUSB, DetectUnit;

type
  TUSBAIConfig = packed record
    ChannelRange: array[$0..$F] of Byte;
    CalibMode: Byte;
    TrigMode: Byte;
    StartStopCh: Byte;
    Oversample: Byte;
    StartStopChEx: Byte;
  end;

const
  NAN: Double = 0/0;

var
  DeviceIndex: LongWord;
  AcqToEdit: TEdit;
  VoltsPerCount, CountsPerVolt: Double;
  bBipolar: Boolean;
  M: Double = 0/0;
  B: Double = 0/0;
  Config: TUSBAIConfig;
  bHardwareCal: Boolean = False;

function IsNAN(D: Double): Boolean;
begin
  Result := StrLIComp(PChar(@D), PChar(@NAN), SizeOf(NAN)) = 0;
end;

function CalibrateCount(C: Word): Word;
begin
  Result := Max(0, Min($FFFF, Round(C * M + B)));
end;

procedure TMainForm.FormCreate(Sender: TObject);
var
  S: TSystemTime;
  DetectForm : TDetectForm;
  Status, PID: LongWord;

begin
  MainPages.TabHeight := 1;
  MainPages.TabWidth := 1;
  MainPages.ActivePageIndex := 0;
  MainPagesChange(nil);
  GetLocalTime(S);
  CalFileEdit.Text := IncludeTrailingBackslash(GetCurrentDir) + Format('%.4d-%.2d-%.2d %.2d-%.2d', [S.wYear, S.wMonth, S.wDay, S.wHour, S.wMinute]) + '.bin';

  DeviceIndex := diOnly;
  Status := QueryDeviceInfo(DeviceIndex, @PID, nil, nil, nil, nil);

  if (Status <> ERROR_SUCCESS) or ((PID and $FEE0) <> $8040) then begin
    DetectForm := TDetectForm.Create(self);
    if DetectForm.ShowModal = mrOk then begin
      DeviceIndex := DetectForm.DeviceIndex;
      DetectForm.Free();
    end
    else begin
      DetectForm.Free();
      Application.Terminate();
      Exit;
    end;
  end;

  Config.StartStopCh := $00;
  Config.StartStopChEx := $00;
  Config.Oversample := 40;

  bHardwareCal := ADC_QueryCal(DeviceIndex) = ERROR_SUCCESS;
  if bHardwareCal then begin
    ADC_SetCal(DeviceIndex, ':1TO1:');

    OverviewInstructionsMemo.Lines.Text := 'System Calibration Utility'#13#13+
                                            'This utility will:'#13+
                                            '1) Clear any calibration of the board.'#13+
                                            '2) Allow you to specify what *SHOULD* be seen, in comparison to what *IS* seen, for both near Zero (offset) and near Full-Scale (Gain) errors.'#13+
                                            '3) Create a new Calibration file that will CAUSE the hardware to return what you''ve specified, instead of the uncalibrated values.'#13+
                                            '4) Display the function used to produce the calibration file.'#13+
                                            #13+
                                            'By performing these steps, you can calibrate MORE than just the board... you calibrate the sensors, the screw terminal accessories, etc.';

    SaveInstructionsMemo.Lines.Text := 'The calibration data you have created must now be saved to a unique filename for later use by your program.'#13+
                                       'Please select the filename you would like to use.  When you later write your code, use this filename in the "ADC_SetCal()" function call.'#13+
                                       'To the left is the function used to produce the calibration file. If you prefer, you can acquire uncalibrated data, and calibrate it in software with this function.';
  end
  else begin
    CalFileEdit.Visible := False;
    BrowseButton.Visible := False;

    OverviewInstructionsMemo.Lines.Text := 'System Calibration Utility'#13#13+
                                            'This utility will:'#13+
                                            '1) Allow you to specify what *SHOULD* be seen, in comparison to what *IS* seen, for both near Zero (offset) and near Full-Scale (Gain) errors.'#13+
                                            '2) Display the function needed to perform the calibration in software, since the detected board does not have hardware calibration.'#13+
                                            #13+
                                            'By performing these steps, you can calibrate MORE than just the board... you calibrate the sensors, the screw terminal accessories, etc.';

    SaveInstructionsMemo.Lines.Text := 'To the left is the function needed to perform the calibration in software.'#13+
                                       'Use this in your code to convert raw counts acquired from the board to calibrated counts.';
  end;

  RangeList.ItemIndex := 1;
  RangeListClick(nil);
  RangeInstructionsMemo.Lines.Text := 'If there is a specific range ("gain setting") you would like to optimize your calibration for, select it here.'#13+
                                      'Be aware, the high-precision gain amplifier used in this product should show little variation in calibration between ranges, and the default range is therefore entirely suitable for most situations.';

  OffsetInstructionsMemo.Lines.Text := 'The two values may not match.  This is normal, and is how we will calculate a file that will cause the data to match, programmatically.'#13#13+
                                        'Once you''ve finished creating and uploading the new calibration data, we''ll run a simple check, to make sure you *THEN* see the correct data displayed.'#13#13+
                                        'Click "Next" when you''re comfortable with the data shown.';

  GainInstructionsMemo.Lines.Text := OffsetInstructionsMemo.Lines.Text;
end;

procedure TMainForm.PrevButtonClick(Sender: TObject);
begin
  MainPages.SelectNextPage(False);
end;

procedure TMainForm.NextButtonClick(Sender: TObject);
var
  F: TFileStream;
  Buf: array of Word;
  I: Integer;

begin
  if MainPages.ActivePage = SaveSheet then begin
    if bHardwareCal then begin
      F := TFileStream.Create(CalFileEdit.Text, fmCreate);
      SetLength(Buf, $10000);
      for I := 0 to $FFFF do
        Buf[I] := CalibrateCount(I)
      ;
      F.Write(Buf[0], $20000);
      F.Free;
      SetLength(Buf, 0);
    end;

    Close;
  end
  else begin
    MainPages.SelectNextPage(True);
  end;
end;

procedure TMainForm.CancelBtnClick(Sender: TObject);
var
  AllowChange: Boolean;

begin
  AllowChange := True;
  MainPagesChanging(nil, AllowChange);
  Close;
end;

procedure TMainForm.MainPagesChange(Sender: TObject);
var
  ArrivingAtPage: TTabSheet;

label
  StartOver;

begin
  StartOver:

  ArrivingAtPage := MainPages.ActivePage;

  if ArrivingAtPage = OverviewSheet then begin
    PrevButton.Enabled := False;
  end
  else if ArrivingAtPage = RangeSheet then begin
  end
  else if ArrivingAtPage = OffsetSheet then begin
    B := NAN;
    AcqToEdit := OffsetObservedEdit;
    AcqTime.Enabled := True;
    AcqTime.OnTimer(nil);
  end
  else if ArrivingAtPage = GainSheet then begin
    if IsNAN(B) then begin
      //Didn't set the offset properly, go do that instead
      MainPages.ActivePage := OffsetSheet;
      goto StartOver;
    end;
    M := NAN;
    AcqToEdit := GainObservedEdit;
    AcqTime.Enabled := True;
    AcqTime.OnTimer(nil);
  end
  else if ArrivingAtPage = ConfirmSheet then begin
    if IsNAN(M) then begin
      //Didn't set the gain properly, go do that instead
      MainPages.ActivePage := GainSheet;
      goto StartOver;
    end;
    AcqToEdit := ConfirmObservedEdit;
    AcqTime.Enabled := True;
    AcqTime.OnTimer(nil);
  end
  else if ArrivingAtPage = SaveSheet then begin
    NextButton.Caption := 'Fi&nish';
    FuncMemo.Lines.Text := 'CalCounts = ' + FloatToStr(Round(M * 1000000) / 1000000) + ' * RawCounts - ' + FloatToStr(Round(-B * 100) / 100);
  end;
end;

procedure TMainForm.MainPagesChanging(Sender: TObject; var AllowChange: Boolean);
var
  LeavingPage: TTabSheet;
  PB: Double;

begin
  LeavingPage := MainPages.ActivePage;

  if LeavingPage = OverviewSheet then begin
    PrevButton.Enabled := True;
  end
  else if LeavingPage = RangeSheet then begin
  end
  else if LeavingPage = OffsetSheet then begin
    AcqTime.Enabled := False;
    try
      B := (StrToFloat(OffsetExpectedEdit.Text) - StrToFloat(OffsetObservedEdit.Text)) * CountsPerVolt;
    except
      B := NAN;
    end;
  end
  else if LeavingPage = GainSheet then begin
    AcqTime.Enabled := False;
    try
      if bBipolar then PB := 32768 - B else PB := -B;
      M := (StrToFloat(GainExpectedEdit.Text) * CountsPerVolt + PB) / (StrToFloat(GainObservedEdit.Text) * CountsPerVolt + PB);
    except
      M := NAN;
    end;
  end
  else if LeavingPage = ConfirmSheet then begin
  end
  else if LeavingPage = SaveSheet then begin
    NextButton.Caption := '&Next';
    //The rest is handled in the Next button itself, since this here doesn't get to specify where it's leaving for.
  end;
end;

procedure TMainForm.BrowseButtonClick(Sender: TObject);
var
  FN: String;

begin
  FN := CalFileEdit.Text;
  SaveEm.InitialDir := ExtractFilePath(FN);
  SaveEm.FileName := ExtractFileName(FN);
  if not SaveEm.Execute then Exit;
  CalFileEdit.Text := SaveEm.FileName;
end;

function GetAveragedImmediate(pSample: PWord): LongWord;
var
  //I: Integer;
  //Tot: Integer;
  ADBuf: array of Word;

begin
  SetLength(ADBuf, 128);
  Result := ADC_GetScan(DeviceIndex, @ADBuf[0]);
  if Result <> ERROR_SUCCESS then Exit;
  pSample^ := ADBuf[0];

  {
  Tot := 0;
  for I := 1 to Wt do begin
    Result := ADC_GetImmediate(DeviceIndex, 0, pSample);
    if Result <> ERROR_SUCCESS then Exit;
    Tot := Tot + pSample^;
  end;
  pSample^ := Round(Tot / Wt);
  }
end;

procedure TMainForm.AcqTimeTimer(Sender: TObject);
var
  Sample: Word;
  Ct: Double;
  Status: LongWord;

begin
  Status := GetAveragedImmediate(@Sample);
  if Status = ERROR_SUCCESS then begin
    Ct := Sample;
    if AcqToEdit = GainObservedEdit then Ct := Ct + B;
    if AcqToEdit = ConfirmObservedEdit then Ct := Ct * M + B;
    if bBipolar then Ct := Ct - 32768;
    AcqToEdit.Text := Format('%.5f', [Ct * VoltsPerCount]);
  end
  else begin
    AcqToEdit.Text := 'Error ' + IntToStr(Status);
  end;
end;

procedure TMainForm.RangeListClick(Sender: TObject);
var
  RangeCode: Byte;
  VRange: Integer;
  CalMinTar, CalMaxTar: Double;

begin
  RangeCode := RangeList.ItemIndex;
  if RangeCode >= 8 then Exit;
  bBipolar := (RangeCode and 1) <> 0;
  VRange := 10;
  if (RangeCode and 1) <> 0 then VRange := VRange * 2;
  if (RangeCode and 2) <> 0 then VRange := VRange div 2;
  if (RangeCode and 4) <> 0 then VRange := VRange div 5;
  VoltsPerCount := VRange / 65536;
  CountsPerVolt := 65536 / VRange;
  B := NAN;
  M := NAN;

  CalMinTar := 1638.4;
  CalMaxTar := 63897.6;
  if bBipolar then begin
    CalMinTar := CalMinTar - 32768;
    CalMaxTar := CalMaxTar - 32768;
  end;
  CalMinTar := CalMinTar * VoltsPerCount;
  CalMaxTar := CalMaxTar * VoltsPerCount;

  OffsetInstructionsLabel.Caption := 'OFFSET: Apply a value to your sensor, which should result in a near-minimum scale (~' + Format('%.2f', [CalMinTar]) + ' volt) reading on Channel 0 Input.';
  GainInstructionsLabel.Caption := 'GAIN: Apply a value to your sensor, which should result in a near-full scale (~' + Format('%.2f', [CalMaxTar]) + ' volt) reading on Channel 0 Input.';

  SEDifRadioClick(nil);
end;

procedure TMainForm.CalPaintPaint(Sender: TObject);
var
  R: TRect;
  I: Integer;

begin
  R := CalPaint.Canvas.ClipRect;
  if R.Top = R.Bottom then R := CalPaint.ClientRect;
  R.Top := R.Top div 16;
  R.Bottom := (R.Bottom - 1) div 16;
  for I := R.Top to R.Bottom do
    CalPaint.Canvas.TextOut(8, I * 16, IntToHex(I, 4) + ' > ' + IntToHex(CalibrateCount(I), 4))
  ;
end;

procedure TMainForm.SEDifRadioClick(Sender: TObject);
var
  ConfigBufSize: LongWord;
  RangeCode: Byte;

begin
  RangeCode := RangeList.ItemIndex;
  if RangeCode >= 8 then RangeCode := $00;

  if SERadio.Checked then
    Config.ChannelRange[0] := RangeCode
  else
    Config.ChannelRange[0] := RangeCode or $08
  ;
  
  ConfigBufSize := SizeOf(Config);
  ADC_SetConfig(DeviceIndex, @Config, ConfigBufSize);
end;

end.
